// server.vala
//
// Copyright (C) 2019 Vivien Kraus <vivien@planete-kraus.eu>
//
//  This program is free software: you can redistribute it and/or
//  modify it under the terms of the GNU General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see
//  <http://www.gnu.org/licenses/>.

namespace Tarot {

	public errordomain ServerError {
		INVALID_EVENT,
		INTERNAL_SERVER_ERROR
	}

	public class ServerGame: Object {
		private Tarot.Game m_game;

		private bool[] m_joined;

		public signal void event_setup (uint destination, size_t n_players, bool with_call);
		public signal void event_deal (uint destination, uint[] cards);
		public signal void event_bid (uint destination, uint bid);
		public signal void event_decl (uint destination, bool decl);
		public signal void event_call (uint destination, uint card);
		public signal void event_dog (uint destination, uint[] cards);
		public signal void event_discard (uint destination, uint[] cards);
		public signal void event_handful (uint destination, uint[] cards);
		public signal void event_play (uint destination, uint card);

		public signal void next (uint who);
		public signal void end (int[] scores);

		public void subscribe (Tarot.Player who) {
			m_joined[who] = true;
			foreach (var e in m_game) {
				unicast_event (e, who);
			}
			// Do not use send_next, otherwise the first player will
			// be notified every time.
			Tarot.Player p_next;
			if (m_game.get_next (out p_next) == Tarot.GameError.OK && p_next == who) {
				next (p_next);
			}
		}

		public void unsubscribe (Tarot.Player who) {
			m_joined[who] = false;
		}

		private static uint[] convert_cards (Tarot.Card[] cards) {
			uint[] ret = new uint[cards.length];
			for (size_t i = 0; i < cards.length; i++) {
				ret[i] = cards[i];
			}
			return ret;
		}

		private void unicast_event (Tarot.GameEvent e, Tarot.Player who) {
			if (!m_joined[who]) {
				return;
			}
			size_t n_players;
			bool with_call;
			Tarot.Player main_player;
			Tarot.Player[] owners;
			Tarot.Bid bid;
			bool decl;
			Tarot.Card card;
			Tarot.Card[] cards;
			if (e.get_setup (out n_players, out with_call) == Tarot.GameEventError.OK) {
				event_setup (who, n_players, with_call);
			} else if (e.get_deal_alloc (out main_player, out cards) == Tarot.GameEventError.OK) {
				assert (main_player == who);
				event_deal (who, convert_cards (cards));
			} else if (e.get_deal_all_alloc (out owners) == Tarot.GameEventError.OK) {
				cards = new Tarot.Card[0];
				for (Tarot.Card c = 0; c < 78; c++) {
					if (owners[c] == who) {
						cards += c;
					}
				}
				event_deal (who, convert_cards (cards));
			} else if (e.get_bid (out bid) == Tarot.GameEventError.OK) {
				event_bid (who, (uint) bid);
			} else if (e.get_decl (out decl) == Tarot.GameEventError.OK) {
				event_decl (who, decl);
			} else if (e.get_call (out card) == Tarot.GameEventError.OK) {
				event_call (who, card);
			} else if (e.get_dog_alloc (out cards) == Tarot.GameEventError.OK) {
				event_dog (who, convert_cards (cards));
			} else if (e.get_discard_alloc (out cards) == Tarot.GameEventError.OK) {
				Tarot.Player taker = 0;
				m_game.get_taker (out taker);
				if (who == taker) {
					if (m_game.get_full_discard_alloc (out cards) == Tarot.GameError.OK) {
						event_discard (who, convert_cards (cards));
					} else {
						assert (false);
					}
				} else {
					if (m_game.get_public_discard_alloc (out cards) == Tarot.GameError.OK) {
						event_discard (who, convert_cards (cards));
					} else {
						assert (false);
					}
				}
			} else if (e.get_handful_alloc (out cards) == Tarot.GameEventError.OK) {
				event_handful (who, convert_cards (cards));
			} else if (e.get_card (out card) == Tarot.GameEventError.OK) {
				event_play (who, (uint) card);
			} else {
				// Unimplemented event type, ignored
			}
		}

		private void send_next () {
			Tarot.Player p_next;
			int[] scores;
			if (m_game.get_next (out p_next) == Tarot.GameError.OK && m_joined[p_next]) {
				next (p_next);
			} else if (m_game.get_scores_alloc (out scores) == Tarot.GameError.OK) {
				end (scores);
			}
		}

		private void broadcast_event (Tarot.GameEvent e) {
			size_t n = m_game.n_players ();
			for (size_t i = 0; i < n; i++) {
				unicast_event (e, (Tarot.Player) i);
			}
		}

		public void add_event (Tarot.GameEvent e, Tarot.Player who) throws ServerError {
			Tarot.Player next;
			if (m_game.get_next (out next) == Tarot.GameError.OK) {
				if (next == who) {
					if (m_game.add_event (e) != Tarot.GameError.OK) {
						throw new ServerError.INVALID_EVENT ("This event is not correct");
					}
					broadcast_event (e);
					Tarot.Card[] dog;
					if (m_game.can_autoreveal_alloc (out dog)) {
						var dog_e = new Tarot.GameEvent.alloc_dog (dog);
						if (m_game.add_event (dog_e) != Tarot.GameError.OK) {
							assert (false);
						}
						broadcast_event (dog_e);
					}
					send_next ();
				} else {
					throw new ServerError.INVALID_EVENT ("This is not your turn");
				}
			} else {
				throw new ServerError.INVALID_EVENT ("Not expecting a player event");
			}
		}

		public void bid (uint bid, uint myself) throws ServerError {
			Tarot.GameEvent e = new Tarot.GameEvent.alloc_bid ((Tarot.Bid) bid);
			add_event (e, myself);
		}

		public void declare (bool slam, uint myself) throws ServerError {
			Tarot.GameEvent e = new Tarot.GameEvent.alloc_decl (slam);
			add_event (e, myself);
		}

		public void call (uint card, uint myself) throws ServerError {
			Tarot.GameEvent e = new Tarot.GameEvent.alloc_call ((Tarot.Card) card);
			add_event (e, myself);
		}

		private static Tarot.Card[] convert_ids (uint[] cards) {
			Tarot.Card[] ret = new Tarot.Card[cards.length];
			for (size_t i = 0; i < cards.length; i++) {
				ret[i] = (Tarot.Card) (cards[i]);
			}
			return ret;
		}

		public void discard (uint[] cards, uint myself) throws ServerError {
			Tarot.GameEvent e = new Tarot.GameEvent.alloc_discard (convert_ids (cards));
			add_event (e, myself);
		}

		public void handful (uint[] cards, uint myself) throws ServerError {
			Tarot.GameEvent e = new Tarot.GameEvent.alloc_handful (convert_ids (cards));
			add_event (e, myself);
		}

		public void play (uint card, uint myself) throws ServerError {
			Tarot.GameEvent e = new Tarot.GameEvent.alloc_card ((Tarot.Card) card);
			add_event (e, myself);
		}

		public ServerGame (size_t n_players, bool with_call) throws ServerError {
			m_game = new Tarot.Game.alloc ();
			m_joined = new bool[n_players];
			for (size_t i = 0; i < n_players; i++) {
				m_joined[i] = false;
			}
			var e = new Tarot.GameEvent.alloc_setup (n_players, with_call);
			if (m_game.add_event (e) != Tarot.GameError.OK) {
				throw new ServerError.INVALID_EVENT ("Wrong players / call combination");
			}
		}

		public bool deal_random (string seed) {
			char[] bytes = new char[seed.length];
			for (size_t i = 0; i < seed.length; i++) {
				bytes[i] = seed[(long) i];
			}
			var e = new Tarot.GameEvent.alloc_deal_all_random (m_game.n_players (), bytes);
			if (m_game.add_event (e) != Tarot.GameError.OK) {
				return false;
			}
			broadcast_event (e);
			send_next ();
			return true;
		}
	}

	[DBus (name = "eu.planete_kraus.Tarot.Server")]
	public class Server: Object {
		private uint m_counter;
		private DBusConnection m_conn;

		public uint create_random (uint n_players, bool with_call, string seed) throws GLib.Error, ServerError {
			var game = new ServerGame (n_players, with_call);
			if (game.deal_random (seed)) {
				var game_id = m_counter++;
				for (size_t i = 0; i < n_players; i++) {
					Seat s = new ServerSeat ((Tarot.Player) i, game);
					uint registration_id = 0;
					try {
						registration_id = m_conn.register_object (
							"/eu/planete_kraus/Tarot/Server/Games/%u/seat/%u"
							.printf (game_id, (uint) i),
							s);
					} catch (IOError e) {
						stderr.printf ("Error while registering a seat: %s\n", e.message);
						throw new ServerError.INTERNAL_SERVER_ERROR ("This has been logged.");
					}
					uint milliseconds = 1;
					uint seconds = 1000 * milliseconds;
					uint minutes = 60 * seconds;
					GLib.Timeout.add (30 * minutes, () => {
							m_conn.unregister_object (registration_id);
							return false;
						});
				}
				return game_id;
			} else {
				stderr.printf ("Error while registering a seat: wrong seed\n");
				throw new ServerError.INTERNAL_SERVER_ERROR ("This has been logged.");
			}
		}

		public Server (DBusConnection conn) {
			m_conn = conn;
		}
	}

	[DBus (name = "eu.planete_kraus.Tarot.Seat")]
	public interface Seat: Object {

		public signal void event_setup (uint n_players, bool with_call);
		public signal void event_deal (uint[] cards);
		public signal void event_bid (uint bid);
		public signal void event_decl (bool decl);
		public signal void event_call (uint card);
		public signal void event_dog (uint[] cards);
		public signal void event_discard (uint[] cards);
		public signal void event_handful (uint[] cards);
		public signal void event_play (uint card);

		public signal void next ();
		public signal void end (int[] scores);

		public abstract uint me () throws GLib.Error;
		public abstract void subscribe () throws GLib.Error;
		public abstract void bid (uint bid) throws GLib.Error, ServerError;
		public abstract void declare (bool slam) throws GLib.Error, ServerError;
		public abstract void call (uint card) throws GLib.Error, ServerError;
		public abstract void discard (uint[] cards) throws GLib.Error, ServerError;
		public abstract void handful (uint[] cards) throws GLib.Error, ServerError;
		public abstract void play (uint card) throws GLib.Error, ServerError;
	}

	public class ServerSeat: Object, Seat {
		private weak ServerGame m_game;
		private Tarot.Player myself;

		public uint me () throws GLib.Error {
			return (uint) myself;
		}

		public void subscribe () throws GLib.Error {
			m_game.subscribe (myself);
		}

		public void unsubscribe () throws GLib.Error {
			m_game.unsubscribe (myself);
		}

		public void bid (uint bid) throws GLib.Error, ServerError {
			m_game.bid (bid, myself);
		}

		public void declare (bool slam) throws GLib.Error, ServerError {
			m_game.declare (slam, myself);
		}

		public void call (uint card) throws GLib.Error, ServerError {
			m_game.call (card, myself);
		}

		public void discard (uint[] cards) throws GLib.Error, ServerError {
			m_game.discard (cards, myself);
		}

		public void handful (uint[] cards) throws GLib.Error, ServerError {
			m_game.handful (cards, myself);
		}

		public void play (uint card) throws GLib.Error, ServerError {
			m_game.play (card, myself);
		}

		public ServerSeat (Tarot.Player p, ServerGame game) {
			m_game = game;
			myself = p;

			game.event_setup.connect ((dest, n_players, with_call) => {
					if (dest == myself) {
						event_setup ((uint) n_players, with_call);
					}
				});
			game.event_deal.connect ((dest, cards) => {
					if (dest == myself) {
						event_deal (cards);
					}
				});
			game.event_bid.connect ((dest, bid) => {
					if (dest == myself) {
						event_bid (bid);
					}
				});
			game.event_decl.connect ((dest, decl) => {
					if (dest == myself) {
						event_decl (decl);
					}
				});
			game.event_call.connect ((dest, call) => {
					if (dest == myself) {
						event_call (call);
					}
				});
			game.event_dog.connect ((dest, cards) => {
					if (dest == myself) {
						event_dog (cards);
					}
				});
			game.event_discard.connect ((dest, cards) => {
					if (dest == myself) {
						event_discard (cards);
					}
				});
			game.event_handful.connect ((dest, cards) => {
					if (dest == myself) {
						event_handful (cards);
					}
				});
			game.event_play.connect ((dest, card) => {
					if (dest == myself) {
						event_play (card);
					}
				});
			game.next.connect ((who) => {
					if (who == myself) {
						next ();
					}
				});
			game.end.connect ((scores) => {
					end (scores);
				});
		}
	}
}
